
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>                    

#include "common.h"

int *xbecz;
int *ybecz;    
int *isbecz;
int ilbecz; 
int *startx;
int *starty;
int *startkier;
int ilstart;    
char mapa[RMP][RMP]; 

void *xmalloc(size_t size)
{
        void *p = malloc(size);
        if (!p) {  
                puts("Out of memory...");
                abort();
        }
        return p;
}

void readmap(const char* file1, const char* file2)
{
	int uchwyt, f;
	FILE *f2;
	if ((uchwyt = open(file1, O_RDWR)) == -1) {
		perror(file1);
		exit(-1);
	}
	if (read(uchwyt, mapa, RMP * RMP) != RMP * RMP) {
		perror(file1);
		exit(-1);
	}
	if (close(uchwyt) == -1) {
		perror(file1);
		exit(-1);
	}
	f2 = fopen(file2, "r");
	if (!f2) {
		perror(file2);
		exit(-1);
	}
	if (1 != fscanf(f2, " ilbecz: %d", &ilbecz) || ilbecz < 0) {
		fprintf(stderr, "le skonstruowany plik %s\n", file2);
		exit(-1);
	}
	xbecz = xmalloc(sizeof(int)*ilbecz);
	ybecz = xmalloc(sizeof(int)*ilbecz);
	isbecz = xmalloc(sizeof(int)*ilbecz);
	for (f = 0; f < ilbecz; f++) {
		if (2 != fscanf(f2, " x: %d y: %d", xbecz+f, ybecz+f)) {
			fprintf(stderr, "le skonstruowany plik %s\n", file2);
			exit(-1);
		}
		isbecz[f] = 1;
	}
	if (1 != fscanf(f2, " ilstart: %d", &ilstart) || ilstart < 1) {
		fprintf(stderr, "le skonstruowany plik %s\n", file2);
		exit(-1);
	}
	startx = xmalloc(sizeof(int)*ilstart);
	starty = xmalloc(sizeof(int)*ilstart);
	startkier = xmalloc(sizeof(int)*ilstart);
	for (f = 0; f < ilstart; f++) {
		if (3 != fscanf(f2, " x: %d y: %d kierunek: %d",
		                startx+f, starty+f, startkier+f)) {
			fprintf(stderr, "le skonstruowany plik %s\n", file2);
			exit(-1);
		}
	}
	if (0 != fclose(f2)) {
		perror(file2);
		exit(-1);
	}
}

void freemap(void)
{
	free(xbecz);
	free(ybecz);
	free(isbecz);
	free(startx);
	free(starty);
	free(startkier);
}

void readconf(Conf * cs, const char *filename)
{
	int f;
	FILE *cf = fopen(filename, "r");
	if (!cf) {
		perror(filename);
		exit(1);
	}
	if (3 != fscanf(cf, " serwer: %as %hd ilegraczy: %d",
			&(cs->serwaddr), &(cs->serwport), &(cs->ilelud))) {
		fprintf(stderr, "le skonstruowany plik %s\n", filename);
		exit(1);
	}
	cs->addrs = (char **) xmalloc(sizeof(char *) * (cs->ilelud));
	cs->ports = (short *) xmalloc(sizeof(short) * (cs->ilelud));
	cs->xs = (int32_t *) xmalloc(sizeof(int32_t) * (cs->ilelud));
	cs->ys = (int32_t *) xmalloc(sizeof(int32_t) * (cs->ilelud));
	cs->kier = (int32_t *) xmalloc(sizeof(int32_t) * (cs->ilelud));

	for (f = 0; f < cs->ilelud; f++) {
		if (5 != fscanf(cf, "%as %hd x: %d y: %d kierunek: %d",
				&(cs->addrs[f]), &(cs->ports[f]),
				&(cs->xs[f]), &(cs->ys[f]),
				&(cs->kier[f]))) {
			fprintf(stderr, "le skonstruowany plik %s (%d)\n",
				filename, f);
			exit(1);
		}
	}
}

void freeconf(Conf * cs)
{
	int f;
	for (f = 0; f < cs->ilelud; f++)
		free(cs->addrs[f]);
	free(cs->serwaddr);
	free(cs->addrs);
	free(cs->ports);
	free(cs->xs);
	free(cs->ys);
	free(cs->kier);
}

