/* 
 * Copyright 2000 Pawel Aleksander Fedorynski
 *
 * You can do with this code whatever you want.
 *
 */

#include <errno.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <arpa/inet.h>
#include <netinet/in.h>
#include <math.h>
#include <stdlib.h>
#include <vga.h>
#include <vgagl.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <vgakeyboard.h>
#include <sys/time.h>
#include <time.h>

#include "common.h"

#define POZIOMO ( jakitryb -> width )
#define PIONOWO ( jakitryb -> height )
#define ILEBAJTOW ( ( jakitryb -> memory )*1024 )
#ifndef XMIN
#define XMIN 11
#endif
#ifndef XMAX
#define XMAX ( POZIOMO-11 )
#endif
#ifndef YMAX
#define YMAX ( PIONOWO-11 )
#endif
#ifndef YMIN
#define YMIN 11
#endif
#define RBMP 54
#define MINCOMP 20
#define TRYB G320x200x256

vga_modeinfo *jakitryb;
GraphicsContext phs;

void pisz(char *tekst)
{
	if (vga_getcurrentmode() == TEXT) {
		printf("\n%s\n", tekst);
		fflush(NULL);
		return;
	} else {
		char *tx = (char *) xmalloc(strlen(tekst) + 1);
		memset(tx, ' ', strlen(tekst));
		tx[strlen(tekst)] = '\0';
		gl_setfont(8, 8, gl_font8x8);
		gl_setwritemode(FONT_COMPRESSED);
		gl_write(XMIN + 1, YMAX - 11, tekst);
		keyboard_close();
		vga_getch();
		gl_write(XMIN + 1, YMAX - 11, tx);
		free(tx);
	}
}

/*
 * Koczy prac.
 */
void koniec()
{
	vga_setmode(TEXT);
	keyboard_close();
	pisz("Koniec.");
	exit(1);
}

/*
 * Wypisuje komunikat o bdzie i koczy prac.
 */
void wrong(char *komunikat, ...)
{
	char *wsk;
	va_list plist;
	vga_getch();
	va_start(plist, komunikat);
	vga_setmode(TEXT);
	pisz("Wystpi bd:");
	pisz(komunikat);
	vga_getch();
	while ((wsk = va_arg(plist, char *))) {
		pisz(wsk);
		vga_getch();
	}
	va_end(plist);
	koniec();
}

/*
 * Element globalnej listy otwartych plikw, z ktrych odczytywane s
 * obiekty typu Bitmapa. Dziki temu odczytujc kolejn bitmap mona
 * po prostu podawa nazw pliku.
 */
typedef struct _nscpl {
	char *path;
	int uchwyt;
	struct _nscpl *sled;
	struct _nscpl *pred;
} nscpl;

/*
 * Kwadratowy obrazek, jeden bajt to jeden piksel.
 */
typedef struct {
	int szer;
	int wys;
	void *zawartosc;
	void *zawmask;
	int dlugmask;
} Bitmapa;

/*
 * Odbija bitmap symetrycznie wzgldem pionowej osi.
 */
void bmp_symlp(Bitmapa * b)
{
	int f, g;
	if (!b->zawartosc)
		return;
	for (f = 0; f < b->wys; f++)
		for (g = 0; g < b->szer / 2; g++) {
			char buf =
			    ((char *) b->zawartosc)[b->szer * f + g];
			((char *) b->zawartosc)[b->szer * f + g]
			    = ((char *) b->zawartosc)[b->szer * f +
						      b->szer - g - 1];
			((char *) b->zawartosc)[b->szer * f + b->szer - g -
						1] = buf;
		}
	free(b->zawmask);
	b->dlugmask =
	    gl_compiledboxmasksize(b->wys, b->szer, b->zawartosc);
	b->zawmask = (char *) xmalloc(b->dlugmask);
	gl_compileboxmask(b->wys, b->szer, b->zawartosc, b->zawmask);
}

void bmp_ini(Bitmapa * b)
{
	b->szer = 0;
	b->wys = 0;
	b->dlugmask = 0;
	b->zawartosc = NULL;
	b->zawmask = NULL;
}

void bmp_free(Bitmapa * b)
{
	free(b->zawartosc);
	free(b->zawmask);
}

/*
 * Kwadratowy obrazek, ktry dodatkowo umie wykonywa pionowe skalowanie.
 * Uywany do wywietlania tekstur na cianach.
 */
typedef struct {
	Bitmapa bb;

	int **skalowanie;
	int own;
} BitmapaSpec;

int bspec_maxskal = 400;

void bspec_free(BitmapaSpec * b)
{
	int f;
	if (b->own) {
		for (f = 0; f < bspec_maxskal; f++)
			free(b->skalowanie[f]);
		free(b->skalowanie);
	}
	bmp_free(&(b->bb));
}

void bspec_ini(BitmapaSpec * b)
{
	bmp_ini(&(b->bb));
	b->skalowanie = NULL;
}

/*
 * Pokazuje jeden przeskalowany pionowy pasek.
 */
void bspec_pp(BitmapaSpec * b, int nrp, int x, int y, int wysp,
	      int hb, int wb, void *vbuf)
{
	int f, yyb, yyk;
	int *skwysp;
	if (wysp >= bspec_maxskal)
		wrong("Za blisko sciany", NULL);
	yyb = (y > 0 ? y : 0);
	yyk = (y + wysp < hb ? y + wysp : hb);
	skwysp = b->skalowanie[wysp];
	for (f = yyb; f < yyk; f++)
		((char *) vbuf)[f * wb + x] =
		    ((char *) b->bb.zawartosc)[nrp + skwysp[f - y]];
}

void bmp_czytaj_fd(Bitmapa * wynik, int uchwyt, int w, int h)
{
	bmp_free(wynik);
	bmp_ini(wynik);
	wynik->szer = w;
	wynik->wys = h;
	wynik->zawartosc = (char *) xmalloc(w * h);
	if (read(uchwyt, wynik->zawartosc, w * h) < w * h) {
		wrong("Nieudane odczytanie bitmapy z pliku.", NULL);
	}
	wynik->dlugmask = gl_compiledboxmasksize(w, h, wynik->zawartosc);
	wynik->zawmask = (char *) xmalloc(wynik->dlugmask);
	gl_compileboxmask(w, h, wynik->zawartosc, wynik->zawmask);
}

nscpl *bitmapa_czyt = NULL;

void bmp_czytaj(Bitmapa * b, char *path, int w, int h)
{
	int uchwyt;
	nscpl *qwerty = bitmapa_czyt;
	while (qwerty) {
		if (!strcmp(qwerty->path, path)) {
			bmp_czytaj_fd(b, qwerty->uchwyt, w, h);
			return;
		}
		qwerty = qwerty->sled;
	}
	if ((uchwyt = open(path, O_RDONLY)) == -1)
		wrong("Nieudane otwarcie pliku ", path, NULL);

	qwerty = (nscpl *) xmalloc(sizeof(nscpl));
	qwerty->uchwyt = uchwyt;
	qwerty->path = (char *) xmalloc(sizeof(char) * (strlen(path) + 1));
	strcpy(qwerty->path, path);
	qwerty->sled = bitmapa_czyt;
	qwerty->pred = NULL;
	if (bitmapa_czyt)
		bitmapa_czyt->pred = qwerty;
	bitmapa_czyt = qwerty;
	bmp_czytaj_fd(b, uchwyt, w, h);
}


void bspec_czytaj(BitmapaSpec * b, char *path, int w, int h)
{
	int f;
	bmp_czytaj(&(b->bb), path, w, h);
	if (b->bb.wys % 2)
		wrong("Nie lubie liczb nieparzystych", NULL);
	b->own = 1;
	b->skalowanie = (int **) xmalloc(sizeof(int *) * (bspec_maxskal));
	for (f = bspec_maxskal; f > 0; f--) {
		b->skalowanie[f - 1] = (int *) xmalloc(sizeof(int) * (f));
		if (f % 2) {
			int g;
			for (g = f / 2; g > 0; g--) {
				int i = f / 2 - g;
				b->skalowanie[f - 1][g] =
				    b->bb.szer * (b->bb.wys / 2 - 1 -
						  i * (b->bb.wys / 2 -
						       1) / (f / 2));
			}
			b->skalowanie[f - 1][0] = 0;
			for (g = f / 2 + 1; g < f - 1; g++) {
				int i = g - f / 2 - 1;
				b->skalowanie[f - 1][g] =
				    b->bb.szer * (b->bb.wys / 2 +
						  i * (b->bb.wys / 2 -
						       1) / (f / 2 - 1));
			}
			b->skalowanie[f - 1][f - 1] =
			    b->bb.szer * (b->bb.wys - 1);
		} else {
			int g;
			for (g = f / 2 - 1; g > 0; g--) {
				int i = f / 2 - 1 - g;
				b->skalowanie[f - 1][g] =
				    b->bb.szer * (b->bb.wys / 2 - 1 -
						  i * (b->bb.wys / 2 -
						       1) / (f / 2 - 1));
			}
			b->skalowanie[f - 1][0] = 0;
			for (g = f / 2; g < f - 1; g++) {
				int i = g - f / 2;
				b->skalowanie[f - 1][g] =
				    b->bb.szer * (b->bb.wys / 2 +
						  i * (b->bb.wys / 2 -
						       1) / (f / 2 - 1));
			}
			b->skalowanie[f - 1][f - 1] =
			    b->bb.szer * (b->bb.wys - 1);
		}
	}
}

/*
 * Sprawdza, czy podany punkt znajduje si poza bitmap.
 */
int bmp_poza(Bitmapa * b, int x, int y)
{
	int retv;
	if (!b->zawartosc)
		retv = 1;
	else if (x >= 0 && x < b->szer && y >= 0 && y < b->wys)
		retv = 0;
	else
		retv = 1;
	return retv;
}

/*
 * Pobiera zawarto jednego bajtu z bitmapy.
 */
char bmp_bajt(Bitmapa * b, int x, int y)
{
	char retv;
	if (bmp_poza(b, x, y))
		retv = 0;
	else
		retv = ((char *) b->zawartosc)[b->szer * y + x];
	return retv;
}

/*
 * Pobiera adres punktu o podanych wsprzdnych w bitmapie.
 */
void *bmp_adr(Bitmapa * b, int x, int y)
{
	if (bmp_poza(b, x, y))
		return NULL;
	return ((char *) b->zawartosc + b->szer * y + x);
}

int bmp_pusta_linia(Bitmapa * b, int linia, int poczatek)
{
	int f;
	int retv;
	if (!b->zawartosc) {
		retv = 1;
		goto koniec;
	}
	if ((linia < 0) || (linia >= b->wys)) {
		retv = 1;
		goto koniec;
	}
	if (poczatek >= b->szer) {
		retv = 1;
		goto koniec;
	}
	for (f = poczatek; f < b->szer; f++)
		if (bmp_bajt(b, f, linia)) {
			retv = 0;
			goto koniec;
		}
	retv = 1;

      koniec:

	return retv;
}

/*
 * Obrazek o (prawie) dowolnym ksztacie. Uywana do pokazywania broni.
 */
typedef struct {
	int pamietanie;
	int ostx;
	int osty;
	GraphicsContext *ostgc;
	int ileblokow;
	int *y0;
	int *liczbaln;
	int **poczln;
	int **dlugln;
	int ilebajtow;
	char *zawartosc;
	char *podemna;
} Klatka;

void klatka_free(Klatka * k)
{
	int f;
	free(k->y0);
	free(k->liczbaln);
	for (f = 0; f < k->ileblokow; f++) {
		free(k->poczln[f]);
		free(k->dlugln[f]);
	}
	free(k->poczln);
	free(k->dlugln);
	free(k->zawartosc);
	free(k->podemna);
}

void klatka_ini(Klatka * k)
{
	k->pamietanie = 1;
	k->ileblokow = 0;
	k->y0 = NULL;
	k->ostx = -1;
	k->osty = -1;
	k->ostgc = NULL;
	k->liczbaln = NULL;
	k->poczln = NULL;
	k->dlugln = NULL;
	k->zawartosc = NULL;
	k->podemna = NULL;
}

int klatka_blokmax = 50;

/*
 * Klatka, ktra posiada dodatkowe informacje umoliwiajce
 * skalowanie obrazka. Uywana do pokazywania graczy, wybuchw, beczek.
 */
typedef struct {
	Klatka kk;
	int szer;
	int wys;
	int **gdziezaczynasieln;
} KlatkaSpec;

void kspec_ini(KlatkaSpec * k)
{
	k->szer = 0;
	k->wys = 0;
	k->gdziezaczynasieln = NULL;
}

/*
 * Funkcja pomocnicza woana przed wywietleniem przeskalowanej klatki.
 * Oblicza wspczynniki skalowania.
 */
void kspec_prep_help(KlatkaSpec * k, int y, int wysp, int hb, int **help)
{
	int f;
	if (wysp == 0) {
		*help = NULL;
		return;
	}
	if (y > hb) {
		*help = NULL;
		return;
	}
	*help = (int *) xmalloc((hb - y + 1) * sizeof(int));
	if (*help == NULL)
		return;
	for (f = 0; f < hb - y; f++)
		(*help)[f] = f * k->wys / wysp;
	(*help)[hb - y] = 0;
}

/*
 * Pokazuje jeden przeskalowany pionowy pasek z klatki.
 */
void kspec_ppq(KlatkaSpec * k, int nrp, int x, int y, int wysp, int hb,
	       int wb, void *vbuf, int *help)
{
	int f, g, linijka;
	if (!help)
		return;
	for (f = 0; f < k->kk.ileblokow; f++) {
		int gmin =
		    k->kk.y0[f] * wysp / k->wys +
		    (k->kk.y0[f] * wysp % k->wys ? 1 : 0);

		int gb = (gmin + y > 0 ? gmin : -y);
		if (gb > hb - y)
			continue;
		for (g = gb, linijka = help[g] - k->kk.y0[f];
		     linijka < k->kk.liczbaln[f] && g + y < hb;
		     g++, linijka = help[g] - k->kk.y0[f])
		{
			if (k->kk.poczln[f][linijka] <= nrp
			    && k->kk.poczln[f][linijka] +
			    k->kk.dlugln[f][linijka] > nrp) {
				((char *) vbuf)[(g + y) * wb + x] =
				    ((char *) k->kk.zawartosc)[k->
							       gdziezaczynasieln
							       [f][linijka]
							       + nrp -
							       k->kk.
							       poczln[f]
							       [linijka]];
			}
		}
	}
}

void kspec_free(KlatkaSpec * k)
{
	int f;
	for (f = 0; f < k->kk.ileblokow; f++)
		free(k->gdziezaczynasieln[f]);
	free(k->gdziezaczynasieln);
}

/*
 * Przygotowuje klatk na podstawie odczytanego z pliku kwadratowego
 * obrazka. Piksele o wartoci 0 (czarne) s traktowane jak przezroczyste.
 */
void klatka_zrob(Klatka * wynik, Bitmapa * bmp)
{
	int f, g, cnb, iloscbajtow, nrobiegu, bwo;
	int *poczatek = (int *) calloc(bmp->wys, sizeof(int));
	int *liczbaln = (int *) xmalloc(sizeof(int) * (klatka_blokmax));
	int *y0 = (int *) xmalloc(sizeof(int) * (klatka_blokmax));
	int **poczln = (int **) xmalloc(sizeof(int *) * (klatka_blokmax));
	int **dlugln;
	char *wsk;
	klatka_ini(wynik);
	for (f = 0; f < klatka_blokmax; f++)
		poczln[f] = (int *) xmalloc(sizeof(int) * (bmp->wys));
	dlugln = (int **) xmalloc(sizeof(int *) * (klatka_blokmax));
	for (f = 0; f < klatka_blokmax; f++)
		dlugln[f] = (int *) xmalloc(sizeof(int) * (bmp->wys));
	cnb = 0;
	iloscbajtow = 0;
	nrobiegu = 0;
	bwo = 0;
	do {
		int linijka = 0;
		nrobiegu++;
		bwo = 0;
		do {
			while (linijka < bmp->wys
			       && bmp_pusta_linia(bmp, linijka,
						  poczatek[linijka]))
				linijka++;
			if (linijka != bmp->wys) {
				y0[cnb] = linijka;
				do {
					int poz1 = poczatek[linijka], poz2;
					while (!bmp_bajt
					       (bmp, poz1, linijka))
						poz1++;
					poz2 = poz1;
					while (bmp_bajt
					       (bmp, ++poz2, linijka));
					poczln[cnb][linijka - y0[cnb]] =
					    poz1;
					dlugln[cnb][linijka - y0[cnb]] =
					    poz2 - poz1;
					poczatek[linijka] = poz2;
					iloscbajtow +=
					    dlugln[cnb][linijka - y0[cnb]];
					linijka++;
				}
				while (linijka < bmp->wys
				       &&
				       !(bmp_pusta_linia
					 (bmp, linijka,
					  poczatek[linijka])));
				liczbaln[cnb] = linijka - y0[cnb];
				cnb++;
				if (cnb > klatka_blokmax - 1)
					wrong("Za maly blokmax.", NULL);
				bwo++;
			}
		}
		while (linijka < bmp->wys);
	}
	while (bwo);
	if (!iloscbajtow) {
		return;
	}
	wynik->ileblokow = cnb;
	wynik->y0 = (int *) xmalloc(sizeof(int) * (cnb));
	bcopy(y0, wynik->y0, cnb * sizeof(int));
	free(y0);
	wynik->liczbaln = (int *) xmalloc(sizeof(int) * (cnb));
	bcopy(liczbaln, wynik->liczbaln, cnb * sizeof(int));
	free(liczbaln);
	wynik->poczln = (int **) xmalloc(sizeof(int *) * (cnb));
	for (f = 0; f < cnb; f++) {
		wynik->poczln[f] =
		    (int *) xmalloc(sizeof(int) * (wynik->liczbaln[f]));
		bcopy(poczln[f], wynik->poczln[f],
		      (wynik->liczbaln[f]) * sizeof(int));
	}
	for (f = 0; f < klatka_blokmax; f++)
		free(poczln[f]);
	free(poczln);
	wynik->dlugln = (int **) xmalloc(sizeof(int *) * (cnb));
	for (f = 0; f < cnb; f++) {
		wynik->dlugln[f] =
		    (int *) xmalloc(sizeof(int) * (wynik->liczbaln[f]));
		bcopy(dlugln[f], wynik->dlugln[f],
		      (wynik->liczbaln[f]) * sizeof(int));
	}
	for (f = 0; f < klatka_blokmax; f++)
		free(dlugln[f]);
	free(dlugln);
	wynik->ilebajtow = iloscbajtow;
	wynik->zawartosc = (char *) xmalloc(iloscbajtow);
	wsk = wynik->zawartosc;
	for (f = 0; f < wynik->ileblokow; f++)
		for (g = 0; g < wynik->liczbaln[f]; g++) {
			bcopy(bmp_adr
			      (bmp, wynik->poczln[f][g], g + wynik->y0[f]),
			      wsk, wynik->dlugln[f][g]);
			wsk += wynik->dlugln[f][g];
		}
	wynik->podemna = NULL;
	wynik->ostx = -1;
	wynik->osty = -1;
	wynik->ostgc = NULL;
	wynik->pamietanie = 1;
	free(poczatek);
}

/*
 * Przygotowuje skalowaln klatk na podstawie odczytanego z pliku
 * kwadratowego obrazka.
 */
void kspec_zrob(KlatkaSpec * k, Bitmapa * bmp)
{
	int f, g, y00, poczlnmin, licznik;
	kspec_ini(k);
	klatka_zrob(&(k->kk), bmp);
	if (!k->kk.zawartosc)
		return;
	y00 = k->kk.y0[0];
	for (f = 0; f < k->kk.ileblokow; f++)
		k->kk.y0[f] -= y00;
	poczlnmin = k->kk.poczln[0][0];
	for (f = 0; f < k->kk.ileblokow; f++)
		for (g = 0; g < k->kk.liczbaln[f]; g++)
			if (k->kk.poczln[f][g] < poczlnmin)
				poczlnmin = k->kk.poczln[f][g];
	for (f = 0; f < k->kk.ileblokow; f++)
		for (g = 0; g < k->kk.liczbaln[f]; g++)
			k->kk.poczln[f][g] -= poczlnmin;
	k->szer = 0;
	for (f = 0; f < k->kk.ileblokow; f++)
		for (g = 0; g < k->kk.liczbaln[f]; g++)
			if (k->szer <
			    k->kk.poczln[f][g] + k->kk.dlugln[f][g])
				k->szer =
				    k->kk.poczln[f][g] +
				    k->kk.dlugln[f][g];
	k->wys = 0;
	for (f = 0; f < k->kk.ileblokow; f++)
		if (k->wys < k->kk.y0[f] + k->kk.liczbaln[f])
			k->wys = k->kk.y0[f] + k->kk.liczbaln[f];
	k->gdziezaczynasieln =
	    (int **) xmalloc(sizeof(int *) * (k->kk.ileblokow));
	licznik = 0;
	for (f = 0; f < k->kk.ileblokow; f++) {
		k->gdziezaczynasieln[f] =
		    (int *) xmalloc(sizeof(int) * (k->kk.liczbaln[f]));
		for (g = 0; g < k->kk.liczbaln[f]; g++) {
			k->gdziezaczynasieln[f][g] = licznik;
			licznik += k->kk.dlugln[f][g];
		}
	}
}

int klatka_przes_y(Klatka * k)
{
	return k->y0 ? k->y0[0] : 0;
}

int klatka_przes_x(Klatka * k)
{
	int f, g, poczlnmin;
	if (!k->zawartosc)
		return 0;
	poczlnmin = k->poczln[0][0];
	for (f = 0; f < k->ileblokow; f++)
		for (g = 0; g < k->liczbaln[f]; g++)
			if (k->poczln[f][g] < poczlnmin)
				poczlnmin = k->poczln[f][g];
	return poczlnmin;
}

/*
 * Wywietla klatk w caoci.
 */
void klatka_pokaz(Klatka * k, int x, int y, void *vbuf, int poziomo,
		  int pionowo)
{
	int f, g;
	int zd = 0;
	char *wsk;
	char *pod;
	wsk = k->zawartosc;
	if (!k->podemna && k->pamietanie)
		k->podemna = (char *) xmalloc(k->ilebajtow);
	pod = k->podemna;
	for (f = 0; f < k->ileblokow; f++)
		for (g = 0; g < k->liczbaln[f]; g++) {
			if (!zd && k->pamietanie)
				bcopy((char *) vbuf +
				      poziomo * (y + k->y0[f] + g) + x +
				      k->poczln[f][g], pod,
				      k->dlugln[f][g]);
			bcopy(wsk,
			      (char *) vbuf + poziomo * (y + k->y0[f] +
							 g) + x +
			      k->poczln[f][g], k->dlugln[f][g]);
			wsk += k->dlugln[f][g];
			pod += k->dlugln[f][g];
		}
}

/*
 * Inicjalizacja grafiki i klawiatury.
 */
int init()
{
	jakitryb = vga_getmodeinfo(TRYB);
	vga_init();
	vga_setmode(TRYB);
	gl_setcontextvga(TRYB);
	gl_getcontext(&phs);
	gl_setwritemode(FONT_COMPRESSED | WRITEMODE_MASKED);
	gl_setfont(8, 8, gl_font8x8);

	return 0;
}

float s[KATY];
float c[KATY];

int xe(int x)
{
	return x * RKL / bok;
}

int ye(int y)
{
	return y * RKL / bok;
}

int X(int x)
{
	return x / bok;
}

int Y(int y)
{
	return y / bok;
}

/*
 * Rzuca promie (jeden z 320) z podanego punktu w podanym kierunku.
 * W przekazanych wskanikach umieszcza informacje
 * a) w ktry blok na mapie trafi promie (to okrela, z ktrej
 *    tekstury pokaemy pionowy pasek)
 * b) jaka jest odlego punktu trafienia od kamery (okrela wysoko paska)
 * c) w ktrym miejscu bloku nastpio trafienie (okrela, ktry pasek
 *    z tekstury pokaza (tzn. np. lewy, rodkowy, z 4/7 od lewej itp.)
 */
void promien(int *Xakt, int *Yakt, float *dist, int *ratio, int a, int x,
	     int y)
{
	int xprzec, yprzec, dhor, dver;
	(*Xakt) = X(x);
	(*Yakt) = Y(y);
	switch (a / KATY_8) {
	case 1:

		dhor = (int) (bok * (c[a] / s[a]));
		dver = (int) (bok * (s[a] / c[a]));
		yprzec = (int) (y + (bok - (x % bok)) * s[a] / c[a]);
		if (Y(yprzec) == Y(y)) {
			(*Xakt)++;

			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = ((yprzec - y) / s[a]);
				(*ratio) = bok - 1 - (yprzec % bok);
				return;
			}
			(*Yakt)++;
			xprzec =
			    x + (int) (((*Yakt) * bok - y) * c[a] / s[a]);

			xprzec =
			    (X(xprzec) > X(x) ? xprzec : (*Xakt) * bok);

			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = (((*Yakt) * bok - y) / s[a]);
				(*ratio) = xprzec % bok;
				return;
			}
		} else {
			xprzec =
			    (int) (x + (bok - (y % bok)) * c[a] / s[a]);
			(*Yakt)++;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = (((*Yakt) * bok - y) / s[a]);
				(*ratio) = xprzec % bok;
				return;
			}
			yprzec -= dver;

		}
		while (1) {
			if (X(xprzec + dhor) == X(xprzec)) {
				xprzec += dhor;
				(*Yakt)++;


				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    (((*Yakt) * bok - y) / s[a]);
					(*ratio) = xprzec % bok;
					return;
				}
			} else {
				yprzec += dver;
				(*Xakt)++;

				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) = ((yprzec - y) / s[a]);
					(*ratio) =
					    bok - 1 - (yprzec % bok);
					return;
				}
				xprzec += dhor;
				(*Yakt)++;

				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    (((*Yakt) * bok - y) / s[a]);
					(*ratio) = xprzec % bok;
					return;
				}

			}
		}

	case 0:

		if (a == 0) {
			do
				(*Xakt)++;
			while (!mapa[(*Xakt)][(*Yakt)]);
			(*ratio) = bok - 1 - y % bok;
			(*dist) = (*Xakt) * bok - x;
			return;
		}
		dhor = (int) (bok * (c[a] / s[a]));
		dver = (int) (bok * (s[a] / c[a]));
		xprzec = (int) (x + (bok - (y % bok)) * c[a] / s[a]);
		if (X(xprzec) == X(x)) {
			(*Yakt)++;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = ((xprzec - x) / c[a]);
				(*ratio) = xprzec % bok;
				return;
			}
			(*Xakt)++;
			yprzec =
			    y + (int) (((*Xakt) * bok - x) * s[a] / c[a]);

			yprzec =
			    (Y(yprzec) > Y(y) ? yprzec : (*Yakt) * bok);

			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = (((*Xakt) * bok - x) / c[a]);
				(*ratio) = bok - 1 - yprzec % bok;
				return;
			}
		} else {
			yprzec =
			    (int) (y + (bok - (x % bok)) * s[a] / c[a]);
			(*Xakt)++;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = (((*Xakt) * bok - x) / c[a]);
				(*ratio) = bok - 1 - yprzec % bok;
				return;
			}
			xprzec -= dhor;
		}
		while (1) {
			if (Y(yprzec + dver) == Y(yprzec)) {
				yprzec += dver;
				(*Xakt)++;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    (((*Xakt) * bok - x) / c[a]);
					(*ratio) = bok - 1 - yprzec % bok;
					return;
				}
			} else {
				xprzec += dhor;
				(*Yakt)++;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) = ((xprzec - x) / c[a]);
					(*ratio) = xprzec % bok;
					return;
				}
				yprzec += dver;
				(*Xakt)++;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    (((*Xakt) * bok - x) / c[a]);
					(*ratio) = bok - 1 - yprzec % bok;
					return;
				}

			}
		}

	case 2:

		dhor = (int) (bok * (c[a] / s[a]));
		dver = -(int) (bok * (s[a] / c[a]));
		if (a == KATY_4) {
			do
				(*Yakt)++;
			while (!mapa[(*Xakt)][(*Yakt)]);
			(*ratio) = x % bok;
			(*dist) = (*Yakt) * bok - y;
			return;
		}
		yprzec = (int) (y - (x % bok) * s[a] / c[a]);
		if (Y(yprzec) == Y(y)) {
			(*Xakt)--;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = ((yprzec - y) / s[a]);
				(*ratio) = yprzec % bok;
				return;
			}
			(*Yakt)++;
			xprzec =
			    x + (int) (((*Yakt) * bok - y) * c[a] / s[a]);

			xprzec =
			    (X(xprzec) <
			     X(x) ? xprzec : ((*Xakt) + 1) * bok - 1);
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = (((*Yakt) * bok - y) / s[a]);
				(*ratio) = xprzec % bok;
				return;
			}
		} else {
			xprzec =
			    (int) (x + (bok - (y % bok)) * c[a] / s[a]);
			(*Yakt)++;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = (((*Yakt) * bok - y) / s[a]);
				(*ratio) = xprzec % bok;
				return;
			}
			yprzec -= dver;
		}
		while (1) {
			if (X(xprzec + dhor) == X(xprzec)) {
				xprzec += dhor;
				(*Yakt)++;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    (((*Yakt) * bok - y) / s[a]);
					(*ratio) = xprzec % bok;
					return;
				}
			} else {
				yprzec += dver;
				(*Xakt)--;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) = ((yprzec - y) / s[a]);
					(*ratio) = yprzec % bok;
					return;
				}
				xprzec += dhor;
				(*Yakt)++;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    (((*Yakt) * bok - y) / s[a]);
					(*ratio) = xprzec % bok;
					return;
				}

			}
		}

	case 3:

		dhor = (int) (bok * (c[a] / s[a]));
		dver = -(int) (bok * (s[a] / c[a]));
		xprzec = (int) (x + (bok - y % bok) * c[a] / s[a]);
		if (X(xprzec) == X(x)) {
			(*Yakt)++;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = ((xprzec - x) / c[a]);
				(*ratio) = xprzec % bok;
				return;
			}
			(*Xakt)--;
			yprzec = y - (int) ((x % bok) * s[a] / c[a]);

			yprzec =
			    (Y(yprzec) > Y(y) ? yprzec : (*Yakt) * bok);
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) =
				    ((((*Xakt) + 1) * bok - x) / c[a]);
				(*ratio) = yprzec % bok;
				return;
			}
		} else {
			yprzec = (int) (y - (x % bok) * s[a] / c[a]);
			(*Xakt)--;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) =
				    ((((*Xakt) + 1) * bok - x) / c[a]);
				(*ratio) = yprzec % bok;
				return;
			}
			xprzec -= dhor;
		}
		while (1) {
			if (Y(yprzec + dver) == Y(yprzec)) {
				yprzec += dver;
				(*Xakt)--;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    ((((*Xakt) + 1) * bok -
					      x) / c[a]);
					(*ratio) = yprzec % bok;
					return;
				}
			} else {
				xprzec += dhor;
				(*Yakt)++;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) = ((xprzec - x) / c[a]);
					(*ratio) = xprzec % bok;
					return;
				}
				yprzec += dver;
				(*Xakt)--;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    ((((*Xakt) + 1) * bok -
					      x) / c[a]);
					(*ratio) = yprzec % bok;
					return;
				}

			}
		}

	case 4:

		dhor = -(int) (bok * (c[a] / s[a]));
		dver = -(int) (bok * (s[a] / c[a]));
		if (a == KATY_2) {
			do
				(*Xakt)--;
			while (!mapa[(*Xakt)][(*Yakt)]);
			(*ratio) = y % bok;
			(*dist) = x - ((*Xakt) + 1) * bok;
			return;
		}
		xprzec = (int) (x - (y % bok) * c[a] / s[a]);
		if (X(xprzec) == X(x)) {
			(*Yakt)--;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = ((xprzec - x) / c[a]);
				(*ratio) = bok - 1 - (xprzec % bok);
				return;
			}
			(*Xakt)--;
			yprzec = y - (int) ((x % bok) * s[a] / c[a]);

			yprzec =
			    (Y(yprzec) <
			     Y(y) ? yprzec : ((*Yakt) + 1) * bok - 1);

			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) =
				    ((((*Xakt) + 1) * bok - x) / c[a]);
				(*ratio) = yprzec % bok;
				return;
			}
		} else {
			yprzec = (int) (y - (x % bok) * s[a] / c[a]);
			(*Xakt)--;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) =
				    ((((*Xakt) + 1) * bok - x) / c[a]);
				(*ratio) = yprzec % bok;
				return;
			}
			xprzec -= dhor;

		}
		while (1) {
			if (Y(yprzec + dver) == Y(yprzec)) {
				yprzec += dver;
				(*Xakt)--;

				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    ((((*Xakt) + 1) * bok -
					      x) / c[a]);
					(*ratio) = yprzec % bok;
					return;
				}
			} else {
				xprzec += dhor;
				(*Yakt)--;

				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) = ((xprzec - x) / c[a]);
					(*ratio) =
					    bok - 1 - (xprzec % bok);
					return;
				}
				yprzec += dver;
				(*Xakt)--;

				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    ((((*Xakt) + 1) * bok -
					      x) / c[a]);
					(*ratio) = yprzec % bok;
					return;
				}

			}
		}

	case 7:

		dhor = -(int) (bok * (c[a] / s[a]));
		dver = (int) (bok * (s[a] / c[a]));
		xprzec = (int) (x - (y % bok) * c[a] / s[a]);
		if (X(xprzec) == X(x)) {
			(*Yakt)--;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = ((xprzec - x) / c[a]);
				(*ratio) = bok - 1 - (xprzec % bok);
				return;
			}
			(*Xakt)++;
			yprzec =
			    y + (int) (((*Xakt) * bok - x) * s[a] / c[a]);

			yprzec =
			    (Y(yprzec) <
			     Y(y) ? yprzec : ((*Yakt) + 1) * bok - 1);

			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = (((*Xakt) * bok - x) / c[a]);
				(*ratio) = bok - 1 - yprzec % bok;
				return;
			}
		} else {
			yprzec = (int) (y + (bok - x % bok) * s[a] / c[a]);
			(*Xakt)++;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = (((*Xakt) * bok - x) / c[a]);
				(*ratio) = bok - 1 - yprzec % bok;
				return;
			}
			xprzec -= dhor;
		}
		while (1) {
			if (Y(yprzec + dver) == Y(yprzec)) {
				yprzec += dver;
				(*Xakt)++;

				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    (((*Xakt) * bok - x) / c[a]);
					(*ratio) = bok - 1 - yprzec % bok;
					return;
				}
			} else {
				xprzec += dhor;
				(*Yakt)--;

				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) = ((xprzec - x) / c[a]);
					(*ratio) =
					    bok - 1 - (xprzec % bok);
					return;
				}
				yprzec += dver;
				(*Xakt)++;

				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    (((*Xakt) * bok - x) / c[a]);
					(*ratio) = bok - 1 - yprzec % bok;
					return;
				}
			}
		}

	case 6:

		dhor = -(int) (bok * (c[a] / s[a]));
		dver = (int) (bok * (s[a] / c[a]));
		if (a == KATY_34) {
			do
				(*Yakt)--;
			while (!mapa[(*Xakt)][(*Yakt)]);
			(*ratio) = bok - 1 - x % bok;
			(*dist) = y - ((*Yakt) + 1) * bok;
			return;
		}
		yprzec = (int) (y + (bok - (x % bok)) * s[a] / c[a]);
		if (Y(yprzec) == Y(y)) {
			(*Xakt)++;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = ((yprzec - y) / s[a]);
				(*ratio) = bok - 1 - (yprzec % bok);
				return;
			}
			(*Yakt)--;
			xprzec = x - (int) ((y % bok) * c[a] / s[a]);

			xprzec =
			    (X(xprzec) > X(x) ? xprzec : (*Xakt) * bok);
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) =
				    ((((*Yakt) + 1) * bok - y) / s[a]);
				(*ratio) = bok - 1 - xprzec % bok;
				return;
			}
		} else {
			xprzec = (int) (x - (y % bok) * c[a] / s[a]);
			(*Yakt)--;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) =
				    ((((*Yakt) + 1) * bok - y) / s[a]);
				(*ratio) = bok - 1 - xprzec % bok;
				return;
			}
			yprzec -= dver;
		}
		while (1) {
			if (X(xprzec + dhor) == X(xprzec)) {
				xprzec += dhor;
				(*Yakt)--;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    ((((*Yakt) + 1) * bok -
					      y) / s[a]);
					(*ratio) = bok - 1 - xprzec % bok;
					return;
				}
			} else {
				yprzec += dver;
				(*Xakt)++;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) = ((yprzec - y) / s[a]);
					(*ratio) =
					    bok - 1 - (yprzec % bok);
					return;
				}
				xprzec += dhor;
				(*Yakt)--;
				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    ((((*Yakt) + 1) * bok -
					      y) / s[a]);
					(*ratio) = bok - 1 - xprzec % bok;
					return;
				}

			}
		}

	case 5:

		dhor = -(int) (bok * (c[a] / s[a]));
		dver = -(int) (bok * (s[a] / c[a]));
		yprzec = (int) (y - (x % bok) * s[a] / c[a]);
		if (Y(yprzec) == Y(y)) {
			(*Xakt)--;

			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) = ((yprzec - y) / s[a]);
				(*ratio) = (yprzec % bok);
				return;
			}
			(*Yakt)--;
			xprzec = x - (int) ((y % bok) * c[a] / s[a]);

			xprzec =
			    (X(xprzec) <
			     X(x) ? xprzec : ((*Xakt) + 1) * bok - 1);

			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) =
				    ((((*Yakt) + 1) * bok - y) / s[a]);
				(*ratio) = bok - 1 - xprzec % bok;
				return;
			}
		} else {
			xprzec = (int) (x - (y % bok) * c[a] / s[a]);
			(*Yakt)--;
			if (mapa[(*Xakt)][(*Yakt)]) {
				(*dist) =
				    ((((*Yakt) + 1) * bok - y) / s[a]);
				(*ratio) = bok - 1 - xprzec % bok;
				return;
			}
			yprzec -= dver;

		}
		while (1) {
			if (X(xprzec + dhor) == X(xprzec)) {
				xprzec += dhor;
				(*Yakt)--;


				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    ((((*Yakt) + 1) * bok -
					      y) / s[a]);
					(*ratio) = bok - 1 - xprzec % bok;
					return;
				}
			} else {
				yprzec += dver;
				(*Xakt)--;

				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) = ((yprzec - y) / s[a]);
					(*ratio) = (yprzec % bok);
					return;
				}
				xprzec += dhor;
				(*Yakt)--;

				if (mapa[(*Xakt)][(*Yakt)]) {
					(*dist) =
					    ((((*Yakt) + 1) * bok -
					      y) / s[a]);
					(*ratio) = bok - 1 - xprzec % bok;
					return;
				}
			}
		}

	default:
		wrong("Not yet...", NULL);

	}
}

/*
 * Rozmiar tekstury.
 */
#define RTX 54

/*
 * Liczba tekstur.
 */
#define ILETX 3

#define MINODL 330
#define MINODL2 330
#define MINBECZ 330
#define MINBS 200
#define ZAKLADKA (KATY_8)

/*
 * Tablica przechowujca tekstury.
 */
BitmapaSpec tekstury[ILETX];

/*
 * Struktura reprezentujca nadajcy si do wywietlenia przedmiot.
 * Oprcz skalowalnej klatki zawiera informacje o rozmiarze przedmiotu
 * w wiecie gry (konieczne, aby wiedzie jaki przeskalowany rozmiar
 * wywietli).
 */
typedef struct {
	KlatkaSpec *klawsk;
	int hswiat;
	int hwewn;
	int wwewn;
	int przesx;
	int przesy;
} Przedmiot;

/*
 * Struktura reprezentujca przedmiot majcy swoje pooenie na mapie
 * (moe to by beczka lub ludzik). W porwnaniu z wersj serwera ma
 * dodatkowo wskanik na obiekt typu Przedmiot, potrzebny do wywietlania
 * danego przedmiotu na ekranie.
 */
typedef struct {
	Przedmiot *prz;
	int32_t xpzl;
	int32_t ypzl;
	float dist;
	char is;
} PrzZlok;

typedef PrzZlok Beczka;

/*
 * Struktura reprezentujca pojedynczego gracza.
 */
typedef struct {
	PrzZlok pzl;

	int16_t punkty;
	int32_t kierunek;
	int krok;
	int krokwbok;
	int katkrok;

	char gunst;
	int gunready;

	char poza;
	Przedmiot *pn[2];
	Przedmiot *pd[2];
	Przedmiot *wsch[2];
	Przedmiot *zach[2];
	Przedmiot *pnwsch[2];
	Przedmiot *pnzach[2];
	Przedmiot *pdzach[2];
	Przedmiot *pdwsch[2];

	Klawisze klaw;
	struct sockaddr_in adres;
	int32_t nrostnad;
	int32_t nrostodb;
} Ludzik;

int kierprom[320];
float depth[320];

int hobr = 200;
int wobr = 320;
void *vbuf;
Przedmiot *wybuch;
Klatka *gun;

int wariant;

int modul(int i)
{
	return (i > 0 ? i : -i);
}

float srodek = (320 / 2) * 3 / 2.0;

int kier_to_ekr(int kierprom)
{
	return (int) (1 +
		      asin((kierprom -
			    srodek) * PI / KATY_2) * 0.666666 * KATY_2 /
		      PI + 320 / 2);
}

float smasin(float dx, float dy, float dist)
{
	float as = asin(dy / dist);
	if (dx < 0) {
		if (as > 0)
			as = PI - as;
		else
			as = -PI - as;
	}
	return as;
}

void ludzik_ini(Ludzik * l)
{
	l->punkty = 0;
	l->nrostnad = 0;
	l->nrostodb = 0;
}

void wpisz_adr(struct sockaddr_in *adres, const char *inetadr, short port)
{
	struct in_addr inadres;

	adres->sin_family = AF_INET;
	adres->sin_port = htons(port);
	if (!inet_aton(inetadr, &inadres)) {
		perror("inet_aton()");
		exit(-1);
	}
	adres->sin_addr = inadres;
}

/*
 * Wybiera, ktra klatka ludzika ma by widoczna zalenie od tego, z ktrej
 * strony na niego patrzymy.
 */
void ludzik_ustaw(Ludzik * l, int alpha)
{
	int odlkat;
	alpha = (alpha - KATY_8 + KATY) % KATY;
	odlkat = (l->kierunek + KATY_8) % KATY - alpha;
	if (odlkat < -KATY_16)
		odlkat += KATY;
	switch (((odlkat + KATY_16) / KATY_8 + 6) % 8) {
	case 0:
		l->pzl.prz = l->pn[(int) l->poza];
		break;
	case 1:
		l->pzl.prz = l->pnwsch[(int) l->poza];
		break;
	case 2:
		l->pzl.prz = l->wsch[(int) l->poza];
		break;
	case 3:
		l->pzl.prz = l->pdwsch[(int) l->poza];
		break;
	case 4:
		l->pzl.prz = l->pd[(int) l->poza];
		break;
	case 5:
		l->pzl.prz = l->pdzach[(int) l->poza];
		break;
	case 6:
		l->pzl.prz = l->zach[(int) l->poza];
		break;
	case 7:
		l->pzl.prz = l->pnzach[(int) l->poza];
		break;
	}
}

int por_prz_zlok_wsk(const void *w1, const void *w2)
{
	PrzZlok **b1 = (PrzZlok **) (w1);
	PrzZlok **b2 = (PrzZlok **) (w2);
	if (b1[0]->dist < b2[0]->dist)
		return 1;
	if (b1[0]->dist > b2[0]->dist)
		return -1;
	return 0;
}

/*
 * Pokazuje przedmiot w odpowiednim punkcie ekranu
 * i przeskalowany zgodnie z zasadami perspektywy.
 */
void przedmiot_pokaz(Przedmiot * p, int xp, int yp, float dist,
		     int x, int y, int alfa)
{
	int f;
	int *help = NULL;
	int dxp = xp - x;
	int dyp = yp - y;
	int katp =
	    ((int) (smasin(dxp, dyp, dist) * KATY_2 / PI) + KATY) % KATY;
	wariant = 0;
	if (alfa <= 3 * KATY_4 - ZAKLADKA / 2) {
		wariant = -1;
		if (katp > alfa - ZAKLADKA / 2
		    && katp - alfa < KATY_4 + ZAKLADKA / 2) {
			int hrp =
			    (int) (100 * p->hswiat /
				   (dist * s[katp - alfa + 240]));
			int wrp = p->wwewn * hrp / p->hwewn;
			int przesekrx = p->przesx * hrp / p->hwewn;
			int przesekry = p->przesy * hrp / p->hwewn;
			int fmp =
			    kier_to_ekr(katp - alfa) - wrp / 2 + przesekrx;
			int fmin = (fmp > 0 ? fmp : 0);
			int wekr = p->klawsk->szer * hrp / p->hwewn;
			int fmax = (fmp + wekr < 320 ? fmp + wekr : 320);
			int yekr =
			    100 -
			    (int) ((p->hswiat -
				    500) * 100 / (dist * s[katp - alfa +
							   240]))
			    + przesekry;
			int hekr = p->klawsk->wys * hrp / p->hwewn;
			wariant = 1;
			kspec_prep_help(p->klawsk, yekr, hekr, hobr,
					&help);
			for (f = fmin; f < fmax; f++) {
				if (depth[f] > dist)
					kspec_ppq(p->klawsk,
						  (f -
						   fmp) * p->wwewn / wrp,
						  f, yekr, hekr, hobr,
						  wobr, vbuf, help);
			}
			free(help);
		}
	} else {
		wariant = -2;
		if (katp > alfa - ZAKLADKA / 2) {
			int hrp =
			    (int) (100 * p->hswiat /
				   (dist * s[katp - alfa + 240]));
			int wrp = p->wwewn * hrp / p->hwewn;
			int przesekrx = p->przesx * hrp / p->hwewn;
			int przesekry = p->przesy * hrp / p->hwewn;
			int fmp =
			    kier_to_ekr(katp - alfa) - wrp / 2 + przesekrx;
			int fmin = (fmp > 0 ? fmp : 0);
			int wekr = p->klawsk->szer * hrp / p->hwewn;
			int fmax = (fmp + wekr < 320 ? fmp + wekr : 320);
			int yekr =
			    100 -
			    (int) ((p->hswiat -
				    500) * 100 / (dist * s[katp - alfa +
							   240]))
			    + przesekry;
			int hekr = p->klawsk->wys * hrp / p->hwewn;
			wariant = 2;
			kspec_prep_help(p->klawsk, yekr, hekr, hobr,
					&help);
			for (f = fmin; f < fmax; f++) {
				if (depth[f] > dist)
					kspec_ppq(p->klawsk,
						  (f -
						   fmp) * p->wwewn / wrp,
						  f, yekr, hekr, hobr,
						  wobr, vbuf, help);
			}
			free(help);
		}
		if (katp < (alfa + KATY_4 + ZAKLADKA / 2) % KATY) {
			int hrp =
			    (int) (100 * p->hswiat /
				   (dist * s[katp - alfa + KATY + 240]));
			int wrp = p->wwewn * hrp / p->hwewn;
			int przesekrx = p->przesx * hrp / p->hwewn;
			int przesekry = p->przesy * hrp / p->hwewn;
			int fmp =
			    kier_to_ekr(katp + KATY - alfa) - wrp / 2 +
			    przesekrx;
			int wekr = p->klawsk->szer * hrp / p->hwewn;
			int fmin = (fmp > 0 ? fmp : 0);
			int fmax = (fmp + wekr < 320 ? fmp + wekr : 320);
			int yekr =
			    100 -
			    (int) ((p->hswiat -
				    500) * 100 / (dist * s[katp - alfa +
							   KATY + 240]))
			    + przesekry;
			int hekr = p->klawsk->wys * hrp / p->hwewn;
			wariant = 3;
			kspec_prep_help(p->klawsk, yekr, hekr, hobr,
					&help);
			for (f = fmin; f < fmax; f++) {
				if (depth[f] > dist)
					kspec_ppq(p->klawsk,
						  (f -
						   fmp) * p->wwewn / wrp,
						  f, yekr, hekr, hobr,
						  wobr, vbuf, help);

			}
			free(help);
		}
	}
}

/*
 * Struktura przechowujca pene informacje o toczcej si grze.
 */
typedef struct {
	uint16_t id;

	int ilepzlok;
	PrzZlok **przwsk;
	int ilelud;
	Ludzik **lud;
	int ilenieruchom;
	PrzZlok **nieruchom;
	int ilemaxstrz;
	char *isstrz;
	int32_t *xstrz;
	int32_t *ystrz;
	int32_t *katstrz;
	int krokstrz;

	int sockuchw;

	struct sockaddr_in adrserw;

	int32_t nrostnad;
	int32_t nrostodb;

	Klawisze klaw;
} StanGry;

/* 
 * Odbiera oczekujce pakiety od serwera z informacjami o stanie gry
 * (ktrymi wypeniana jest przekazana struktura). Caa logika gry
 * rozstrzyga si na serwerze.
 */
int gra_sluchaj(StanGry * g)
{
	int f;
	int changed = 0;
	int32_t nrmsg;
	int ileb;
	int msgsize =
	    4 + 13 * g->ilemaxstrz + 17 * g->ilelud + g->ilenieruchom;
	char *sgmsg = (char *) xmalloc(sizeof(char) * (msgsize));
	while ((ileb = recv(g->sockuchw, sgmsg, msgsize, 0)) != -1) {
		char *p = sgmsg;
		int32_t tmp32;
		int16_t tmp16;
		if (ileb != msgsize) {
			puts("WRONG1");
			continue;
		}
		bcopy(p, &nrmsg, 4);
		p += 4;
		nrmsg = ntohl(nrmsg);
		if (g->nrostodb >= nrmsg)
			continue;
		changed = 1;
		g->nrostodb = nrmsg;
		bcopy(p, g->isstrz, g->ilemaxstrz);
		p += g->ilemaxstrz;
		for (f = 0; f < g->ilemaxstrz; f++) {
			bcopy(p, &tmp32, 4);
			g->xstrz[f] = ntohl(tmp32);
			p += 4;
			bcopy(p, &tmp32, 4);
			g->ystrz[f] = ntohl(tmp32);
			p += 4;
			bcopy(p, &tmp32, 4);
			g->katstrz[f] = ntohl(tmp32);
			p += 4;
		}
		for (f = 0; f < g->ilelud; f++) {
			bcopy(p, &tmp16, 2);
			g->lud[f]->punkty = ntohs(tmp16);
			p += 2;
			bcopy(p, &tmp32, 4);
			g->lud[f]->pzl.xpzl = ntohl(tmp32);
			p += 4;
			bcopy(p, &tmp32, 4);
			g->lud[f]->pzl.ypzl = ntohl(tmp32);
			p += 4;
			bcopy(p, &tmp32, 4);
			g->lud[f]->kierunek = ntohl(tmp32);
			p += 4;
			g->lud[f]->pzl.is = p[0];
			g->lud[f]->gunst = p[1];
			g->lud[f]->poza = p[2];
			p += 3;
		}
		for (f = 0; f < g->ilenieruchom; f++) {
			g->nieruchom[f]->is = *p;
			p++;
		}
	}
	free(sgmsg);
	return changed;
}

/*
 * Wysya do serwera stan klawiatury.
 */
void gra_slij_klaw(StanGry * g)
{
	int32_t tmp32;
	uint16_t tmp16;

	int sizeklms = sizeof(Klawisze) + 6;
	char *klawmsg = (char *) xmalloc(sizeof(char) * (sizeklms));
	tmp16 = htons(g->id);
	bcopy(&tmp16, klawmsg, 2);
	g->nrostnad++;
	tmp32 = htonl(g->nrostnad);
	bcopy(&tmp32, klawmsg + 2, 4);
	bcopy(&g->klaw, klawmsg + 6, sizeof(Klawisze));
	sendto(g->sockuchw, klawmsg, sizeklms, 0,
	       (struct sockaddr *) (&g->adrserw), sizeof(g->adrserw));
	free(klawmsg);
}

void gra_sock_init(StanGry * g, const char *inetadr, short port)
{
	struct sockaddr_in adres;
	struct in_addr inadres;

	g->sockuchw = socket(PF_INET, SOCK_DGRAM, 0);
	if (g->sockuchw == -1) {
		perror("socket()");
		exit(-1);
	}
	adres.sin_family = AF_INET;
	adres.sin_port = htons(port);
	if (!inet_aton(inetadr, &inadres)) {
		perror("inet_aton()");
		exit(-1);
	}
	adres.sin_addr = inadres;
	if (-1 ==
	    bind(g->sockuchw, (struct sockaddr *) &adres, sizeof(adres))) {
		perror("bind()");
		exit(-1);
	}
	if (-1 == fcntl(g->sockuchw, F_SETFL, O_NONBLOCK)) {
		perror("fcntl()");
		exit(-1);
	}
}

/*
 * Rysuje stan gry widziany oczami gracza o numerze nrlud.
 */
void gra_rysuj(StanGry * g, int nrlud)
{
	int f;
	int x = g->lud[nrlud]->pzl.xpzl;
	int y = g->lud[nrlud]->pzl.ypzl;
	int alfa = g->lud[nrlud]->kierunek;
	int Xk, Yk, ratio;
	float dist;
	int hrel, hrflag;

	memset(vbuf, 23, hobr * wobr / 2);
	memset((char *) vbuf + hobr * wobr / 2, 20,
	       hobr * wobr - hobr * wobr / 2);

	if (alfa == 240 || alfa == 1680 || alfa == 1200 || alfa == 720) {
		int lastXk = -1;
		int lastYk = -1;
		int lasthrel = 0;
		for (f = 0; f < wobr; f++) {
			promien(&Xk, &Yk, &dist, &ratio,
				(alfa + kierprom[f]) % KATY, x, y);
			depth[f] = dist;
			hrel =
			    (int) (100000 / (dist * s[kierprom[f] + 240]));
			if (Xk == lastXk && Yk == lastYk
			    && modul(hrel - lasthrel) < 2)
				hrel = lasthrel;
			else {
				lasthrel = hrel;
				lastXk = Xk;
				lastYk = Yk;
			}
			hrflag = (hrel > hobr ? 0 : 1);
			bspec_pp(tekstury + mapa[Xk][Yk] - 1,
				 ratio * RTX / bok, f,
				 (hobr - hrel + hrflag) / 2, hrel, hobr,
				 wobr, vbuf);
		}
	} else
		for (f = 0; f < wobr; f++) {
			promien(&Xk, &Yk, &dist, &ratio,
				(alfa + kierprom[f]) % KATY, x, y);
			depth[f] = dist;
			hrel =
			    (int) (100000 / (dist * s[kierprom[f] + 240]));
			hrflag = (hrel > hobr ? 0 : 1);
			bspec_pp(tekstury + mapa[Xk][Yk] - 1,
				 ratio * RTX / bok, f,
				 (hobr - hrel + hrflag) / 2, hrel, hobr,
				 wobr, vbuf);
		}

	qsort(g->przwsk, g->ilepzlok, sizeof(PrzZlok *), por_prz_zlok_wsk);

	for (f = 0; f < g->ilepzlok; f++) {
		if (g->przwsk[f] == (PrzZlok *) (g->lud[nrlud]))
			continue;
		if (g->przwsk[f][0].is > 0) {
			przedmiot_pokaz(g->przwsk[f][0].prz,
					g->przwsk[f][0].xpzl,
					g->przwsk[f][0].ypzl,
					g->przwsk[f][0].dist, x, y, alfa);
		} else {
			if (g->przwsk[f][0].is < 0) {
				przedmiot_pokaz(wybuch + 16 +
						g->przwsk[f][0].is,
						g->przwsk[f][0].xpzl,
						g->przwsk[f][0].ypzl,
						g->przwsk[f][0].dist, x, y,
						alfa);
			}
		}
	}

	klatka_pokaz(gun + g->lud[nrlud]->gunst, wobr / 2 - 32, hobr - 64,
		     vbuf, wobr, hobr);

	bcopy(vbuf, graph_mem, 64000);
}

/*
 * Oblicza odlegoci wszystkich przedmiotw od gracza o numerze nrlud.
 */
void gra_ust_dist(StanGry * g, int nrlud)
{
	int x = g->lud[nrlud]->pzl.xpzl;
	int y = g->lud[nrlud]->pzl.ypzl;
	int f;

	for (f = 0; f < g->ilepzlok; f++)
		g->przwsk[f][0].dist =
		    sqrt((double)
			 ((x - g->przwsk[f][0].xpzl) * (x -
							g->przwsk[f][0].
							xpzl) + (y -
								 g->
								 przwsk[f]
								 [0].
								 ypzl) *
			  (y - g->przwsk[f][0].ypzl)));
}

/*
 * Inicjalizuje struktur StanGry na podstawie odczytanych z pliku
 * konfiguracyjnego ustawie.
 */
void gra_init(StanGry * g, Conf * cs)
{
	int f;
	KlatkaSpec *beczklwsk = (KlatkaSpec *) xmalloc(sizeof(KlatkaSpec));
	Przedmiot *beczwsk = (Przedmiot *) xmalloc(sizeof(Przedmiot));
	KlatkaSpec *ludzikkl =
	    (KlatkaSpec *) xmalloc(sizeof(KlatkaSpec) * (16));
	Przedmiot *ludzikprz =
	    (Przedmiot *) xmalloc(sizeof(Przedmiot) * (16));
	Bitmapa tymcz;
	bmp_ini(&tymcz);

	g->ilelud = cs->ilelud;
	g->ilepzlok = ilbecz + g->ilelud;
	g->przwsk =
	    (PrzZlok **) xmalloc(sizeof(PrzZlok *) * (g->ilepzlok));
	g->ilemaxstrz = 20;
	g->isstrz = (char *) xmalloc(sizeof(char) * (g->ilemaxstrz));
	g->xstrz = (int32_t *) xmalloc(sizeof(int32_t) * (g->ilemaxstrz));
	g->ystrz = (int32_t *) xmalloc(sizeof(int32_t) * (g->ilemaxstrz));
	g->katstrz =
	    (int32_t *) xmalloc(sizeof(int32_t) * (g->ilemaxstrz));
	g->krokstrz = 100;

	memset(g->isstrz, 0, g->ilemaxstrz);

	bmp_czytaj(&tymcz, "rysunki/beczka.rys", 54, 54);
	kspec_zrob(beczklwsk, &tymcz);

	beczwsk->klawsk = beczklwsk;
	beczwsk->hswiat = 500;
	beczwsk->hwewn = beczwsk->klawsk->wys;
	beczwsk->wwewn = beczwsk->klawsk->szer;
	beczwsk->przesx = 0;
	beczwsk->przesy = 0;

	for (f = 0; f < ilbecz; f++) {
		g->przwsk[f] = (Beczka *) xmalloc(sizeof(Beczka));
		g->przwsk[f][0].prz = beczwsk;
		g->przwsk[f][0].xpzl = xbecz[f];
		g->przwsk[f][0].ypzl = ybecz[f];
		g->przwsk[f][0].is = 1;
	}

	g->ilenieruchom = ilbecz;
	g->nieruchom =
	    (PrzZlok **) xmalloc(sizeof(PrzZlok *) * (g->ilenieruchom));
	bcopy(g->przwsk, g->nieruchom,
	      g->ilenieruchom * sizeof(PrzZlok *));

	bmp_czytaj(&tymcz, "rysunki/lud.rys", 64, 64);
	kspec_zrob(ludzikkl + 0, &tymcz);
	bmp_czytaj(&tymcz, "rysunki/lud.rys", 64, 64);
	kspec_zrob(ludzikkl + 1, &tymcz);
	bmp_czytaj(&tymcz, "rysunki/lud.rys", 64, 64);
	kspec_zrob(ludzikkl + 2, &tymcz);
	bmp_symlp(&tymcz);
	kspec_zrob(ludzikkl + 4, &tymcz);
	bmp_czytaj(&tymcz, "rysunki/lud.rys", 64, 64);
	kspec_zrob(ludzikkl + 3, &tymcz);
	bmp_symlp(&tymcz);
	kspec_zrob(ludzikkl + 5, &tymcz);
	bmp_czytaj(&tymcz, "rysunki/lud.rys", 64, 64);
	kspec_zrob(ludzikkl + 6, &tymcz);
	bmp_symlp(&tymcz);
	kspec_zrob(ludzikkl + 8, &tymcz);
	bmp_czytaj(&tymcz, "rysunki/lud.rys", 64, 64);
	kspec_zrob(ludzikkl + 7, &tymcz);
	bmp_symlp(&tymcz);
	kspec_zrob(ludzikkl + 9, &tymcz);
	bmp_czytaj(&tymcz, "rysunki/lud.rys", 64, 64);
	kspec_zrob(ludzikkl + 10, &tymcz);
	bmp_symlp(&tymcz);
	kspec_zrob(ludzikkl + 12, &tymcz);
	bmp_czytaj(&tymcz, "rysunki/lud.rys", 64, 64);
	kspec_zrob(ludzikkl + 11, &tymcz);
	bmp_symlp(&tymcz);
	kspec_zrob(ludzikkl + 13, &tymcz);
	bmp_czytaj(&tymcz, "rysunki/lud.rys", 64, 64);
	kspec_zrob(ludzikkl + 14, &tymcz);
	bmp_czytaj(&tymcz, "rysunki/lud.rys", 64, 64);
	kspec_zrob(ludzikkl + 15, &tymcz);

	for (f = 0; f < 16; f++) {
		ludzikprz[f].klawsk = &(ludzikkl[f]);
		ludzikprz[f].hswiat = 700;
		ludzikprz[f].hwewn = ludzikprz[f].klawsk->wys;
		ludzikprz[f].wwewn = ludzikprz[f].klawsk->szer;
		ludzikprz[f].przesx = 0;
		ludzikprz[f].przesy = 0;
	}

	g->lud = (Ludzik **) xmalloc(sizeof(Ludzik *) * (g->ilelud));

	for (f = 0; f < g->ilelud; f++) {
		g->lud[f] = (Ludzik *) xmalloc(sizeof(Ludzik));
		ludzik_ini(g->lud[f]);

		g->lud[f]->pd[0] = ludzikprz;
		g->lud[f]->pd[1] = ludzikprz + 1;
		g->lud[f]->zach[0] = ludzikprz + 2;
		g->lud[f]->zach[1] = ludzikprz + 3;
		g->lud[f]->wsch[0] = ludzikprz + 4;
		g->lud[f]->wsch[1] = ludzikprz + 5;
		g->lud[f]->pdzach[0] = ludzikprz + 6;
		g->lud[f]->pdzach[1] = ludzikprz + 7;
		g->lud[f]->pdwsch[0] = ludzikprz + 8;
		g->lud[f]->pdwsch[1] = ludzikprz + 9;
		g->lud[f]->pnzach[0] = ludzikprz + 10;
		g->lud[f]->pnzach[1] = ludzikprz + 11;
		g->lud[f]->pnwsch[0] = ludzikprz + 12;
		g->lud[f]->pnwsch[1] = ludzikprz + 13;
		g->lud[f]->pn[0] = ludzikprz + 14;
		g->lud[f]->pn[1] = ludzikprz + 15;
		g->lud[f]->pzl.is = 1;
		g->lud[f]->poza = 0;
		g->lud[f]->gunst = 0;
		g->lud[f]->gunready = 1;
		g->lud[f]->krok = 80;
		g->lud[f]->krokwbok = 25;
		g->lud[f]->katkrok = 10;
	}

	for (f = 0; f < g->ilelud; f++) {
		g->lud[f]->pzl.xpzl = cs->xs[f];
		g->lud[f]->pzl.ypzl = cs->ys[f];
		g->lud[f]->kierunek = cs->kier[f];
	}

	for (f = 0; f < g->ilelud; f++)
		g->przwsk[ilbecz + f] = &(g->lud[f]->pzl);

	bmp_free(&tymcz);
}

void ust_klaw_zw(Klawisze * klaw)
{
	keyboard_update();
	if (keyboard_keypressed(SCANCODE_COMMA))
		klaw->rightstr = 1;
	else
		klaw->rightstr = 0;
	if (keyboard_keypressed(SCANCODE_PERIOD))
		klaw->leftstr = 1;
	else
		klaw->leftstr = 0;
	if (keyboard_keypressed(SCANCODE_CURSORLEFT))
		klaw->left = 1;
	else
		klaw->left = 0;
	if (keyboard_keypressed(SCANCODE_CURSORRIGHT))
		klaw->right = 1;
	else
		klaw->right = 0;
	if (keyboard_keypressed(SCANCODE_CURSORUP))
		klaw->up = 1;
	else
		klaw->up = 0;
	if (keyboard_keypressed(SCANCODE_CURSORDOWN))
		klaw->down = 1;
	else
		klaw->down = 0;
	if (keyboard_keypressed(SCANCODE_CONTROL))
		klaw->fire = 1;
	else
		klaw->fire = 0;
	if (keyboard_keypressed(SCANCODE_RIGHTALT))
		klaw->str = 1;
	else
		klaw->str = 0;
}

int main(int argc, char **argv)
{
	char info[40];
	int f, licznikkl = 0, klpers;
	time_t czas, oldczas;
	Conf cs;
	StanGry gra;
	Bitmapa tymcz;
	bmp_ini(&tymcz);
	vbuf = xmalloc(wobr * hobr);
	if (argc != 2) {
		fprintf(stderr, "Uycie: klient <numer gracza>\n");
		exit(1);
	}
	for (f = 0; f < KATY; f++) {
		s[f] = sin(f * PI / KATY_2);
		c[f] = cos(f * PI / KATY_2);
	}
	for (f = 0; f < 320; f++) {
		kierprom[f] =
		    (int) (srodek +
			   sin((f -
				320 / 2) * 1.5 * PI / KATY_2) * KATY_2 /
			   PI);
	}
	readmap("mapa/m2.mapa", "mapa/m2.pos");
	wybuch = (Przedmiot *) xmalloc(sizeof(Przedmiot) * (16));
	for (f = 0; f < 16; f++) {
		Klatka tmp;
		bmp_czytaj(&tymcz, "rysunki/flaki2.rys", 200, 200);
		wybuch[f].klawsk =
		    (KlatkaSpec *) xmalloc(sizeof(KlatkaSpec));
		kspec_zrob(wybuch[f].klawsk, &tymcz);
		klatka_zrob(&tmp, &tymcz);
		wybuch[f].przesx = klatka_przes_x(&tmp) + 12;
		wybuch[f].przesy = klatka_przes_y(&tmp) + 80;
		wybuch[f].wwewn = 200;
		wybuch[f].hwewn = 200;
		wybuch[f].hswiat = 2000;
		klatka_free(&tmp);
	}
	gun = (Klatka *) xmalloc(sizeof(Klatka) * (12));
	for (f = 0; f < 12; f++) {
		bmp_czytaj(&tymcz, "rysunki/gun.rys", 64, 64);
		klatka_zrob(gun + f, &tymcz);
	}
	for (f = 0; f < ILETX; f++) {
		bspec_ini(tekstury + f);
		bspec_czytaj(tekstury + f, "tekstury/wall.tks", RTX, RTX);
	}
	readconf(&cs, "config");
	gra.nrostnad = 0;
	gra.nrostodb = 0;
	gra.id = atoi(argv[1]);
	if (gra.id >= cs.ilelud) {
		fprintf(stderr,
			"Numer gracza musi by w zakresie 0...%d\n",
			cs.ilelud - 1);
		exit(1);
	}
	gra_init(&gra, &cs);
	gra_sock_init(&gra, cs.addrs[gra.id], cs.ports[gra.id]);
	wpisz_adr(&(gra.adrserw), cs.serwaddr, cs.serwport);

	oldczas = time(NULL);
	init();
	keyboard_init();
	keyboard_translatekeys(TRANSLATE_CURSORKEYS);
	do {
		if (!gra_sluchaj(&gra))
			continue;
		ust_klaw_zw(&(gra.klaw));
		gra_slij_klaw(&gra);
		if (0 >= gra.lud[gra.id]->pzl.is) {
			char komm[100];
			sprintf(komm, "You are dead");
			gl_write(XMIN + 1, YMAX - 100, komm);
			continue;
		}
		gra_ust_dist(&gra, gra.id);
		for (f = 0; f < gra.ilelud; f++) {
			if (f == gra.id)
				continue;
			ludzik_ustaw(gra.lud[f],
				     gra.lud[gra.id]->kierunek);
		}

		gra_rysuj(&gra, gra.id);
		usleep(30000);

		licznikkl++;
		czas = time(NULL);
		if (czas != oldczas) {
			klpers = licznikkl;
			licznikkl = 0;
			oldczas = czas;
		}
		if (keyboard_keypressed(SCANCODE_T)) {
			sprintf(info, "%d  %d", klpers, wariant);
			gl_write(XMIN + 1, YMAX - 11, info);
		}
	}
	while (!keyboard_keypressed(SCANCODE_Q));
	keyboard_close();
	freeconf(&cs);
	bmp_free(&tymcz);
	for (f = 0; f < ILETX; f++)
		bspec_free(tekstury + f);
	freemap();
	pisz("OK");
	return 0;
}
