/*
 * Copyright 2000 Pawel Aleksander Fedorynski
 *
 * You can do with this code whatever you want.
 *
 */

#define DEBUG 0

/*
 * Minimalne odlegoci od cian, beczek itp.
 */
#define MINODL 330
#define MINODL2 330
#define MINBECZ 330
#define MINBS 200
#define ZAKLADKA (KATY_8)

#include <math.h>
#include <errno.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <stdarg.h>
#include <string.h>
#include <sys/time.h>
#include <time.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#include "common.h"

/*
 * Tablice sinusw i cosinusw.
 */
float s[KATY];
float c[KATY];

void klawisze_ini(Klawisze * k)
{
	k->left = 0;
	k->right = 0;
	k->up = 0;
	k->down = 0;
	k->leftstr = 0;
	k->rightstr = 0;
	k->fire = 0;
	k->str = 0;
}

/*
 * Struktura przechowujca dane obiektu majcego swoje pooenie na mapie.
 * Moe to by beczka lub gracz.
 */
typedef struct {
	int32_t xpzl;
	int32_t ypzl;
	float dist;
	char is;
} PrzZlok;

typedef PrzZlok Beczka;

/*
 * Struktura reprezentujca pojedynczego gracza.
 */
typedef struct {
	PrzZlok pzl;

	int16_t punkty;
	int32_t kierunek;
	int krok;
	int krokwbok;
	int katkrok;

	char gunst;
	int gunready;

	char poza;

	Klawisze klaw;
	struct sockaddr_in adres;
	int32_t nrostnad;
	int32_t nrostodb;
} Ludzik;

void ludzik_ini(Ludzik * l)
{
	l->punkty = 0;
	l->nrostnad = 0;
	l->nrostodb = 0;
	klawisze_ini(&(l->klaw));
}

void wpisz_adr(struct sockaddr_in *adres, const char *inetadr, short port)
{
	struct in_addr inadres;

	adres->sin_family = AF_INET;
	adres->sin_port = htons(port);
	if (!inet_aton(inetadr, &inadres)) {
		perror("inet_aton()");
		exit(-1);
	}
	adres->sin_addr = inadres;
}

void ludzik_rusz_gun(Ludzik * l)
{
	if (l->gunst) {
		l->gunst++;
		if (l->gunst == 12)
			l->gunst = 0;
	}
}

typedef struct {
	uint16_t id;

	int ilepzlok;
	PrzZlok **przwsk;
	int ilelud;
	Ludzik **lud;
	int ilenieruchom;
	PrzZlok **nieruchom;
	int ilemaxstrz;
	char *isstrz;
	int32_t *xstrz;
	int32_t *ystrz;
	int32_t *katstrz;
	int krokstrz;

	int oststart;

	int sockuchw;
	struct sockaddr_in adrserw;

	int32_t nrostnad;
	int32_t nrostodb;

	Klawisze klaw;
} StanGry;

/*
 * Procedura woana, gdy gracz ktry by zabity nacinie guzik 'strza'.
 * Ustawia gracza w jednym z odczytanych na mapie punktw startowych.
 */
void gra_fresh_start(StanGry * g, int nrlud)
{
	Klawisze klawzero;
	g->oststart++;
	if (g->oststart == ilstart)
		g->oststart = 0;
	g->lud[nrlud]->pzl.is = 1;
	g->lud[nrlud]->poza = 0;
	g->lud[nrlud]->gunst = 0;
	g->lud[nrlud]->gunready = 1;
	g->lud[nrlud]->pzl.xpzl = startx[g->oststart];
	g->lud[nrlud]->pzl.ypzl = starty[g->oststart];
	g->lud[nrlud]->kierunek = startkier[g->oststart];
	klawisze_ini(&klawzero);
	g->lud[nrlud]->klaw = klawzero;
}

/*
 * Wysya pakiet do gracza nr nrlud.
 */
void gra_slij_do(StanGry * g, int nrlud)
{
	int32_t tmp32;
	int16_t tmp16;
	int msgsize =
	    4 + 13 * g->ilemaxstrz + 17 * g->ilelud + g->ilenieruchom;
	char *sgmsg = (char *) xmalloc(msgsize);
	char *p = sgmsg;
	int f;
	g->lud[nrlud]->nrostnad++;
	tmp32 = htonl(g->lud[nrlud]->nrostnad);
	bcopy(&tmp32, p, 4);
	p += 4;			/*  4 */
	bcopy(g->isstrz, p, g->ilemaxstrz);
	p += g->ilemaxstrz;	/*  4+g->ilemaxstrz */
	for (f = 0; f < g->ilemaxstrz; f++) {
		tmp32 = htonl(g->xstrz[f]);
		bcopy(&tmp32, p, 4);
		p += 4;
		tmp32 = htonl(g->ystrz[f]);
		bcopy(&tmp32, p, 4);
		p += 4;
		tmp32 = htonl(g->katstrz[f]);
		bcopy(&tmp32, p, 4);
		p += 4;
	}			/*  4+g->ilemaxstrz*13 */
	for (f = 0; f < g->ilelud; f++) {
		tmp16 = htons(g->lud[f]->punkty);
		bcopy(&tmp16, p, 2);
		p += 2;
		tmp32 = htonl(g->lud[f]->pzl.xpzl);
		bcopy(&tmp32, p, 4);
		p += 4;
		tmp32 = htonl(g->lud[f]->pzl.ypzl);
		bcopy(&tmp32, p, 4);
		p += 4;
		tmp32 = htonl(g->lud[f]->kierunek);
		bcopy(&tmp32, p, 4);
		p += 4;
		*p = g->lud[f]->pzl.is;
		p++;
		*p = g->lud[f]->gunst;
		p++;
		*p = g->lud[f]->poza;
		p++;
	}			/*  4+13*ilemaxstrz+17*ilelud */
	for (f = 0; f < g->ilenieruchom; f++) {
		*p = g->nieruchom[f]->is;
		p++;
	}			/*  4+13*ilemaxstrz+15*ilelud+ilenieruchom */
	sendto(g->sockuchw, sgmsg, msgsize, 0,
	       (struct sockaddr *) (&(g->lud[nrlud]->adres)),
	       sizeof(g->lud[nrlud]->adres));
	free(sgmsg);
}

/*
 * Odbiera oczekujce, nadesane przez klientw, pakiety z informacjami
 * o klawiszach wciskanych przez graczy.
 */
void gra_sluchaj_klaw(StanGry * g)
{
	int32_t nrkom;
	uint16_t nrlud;

	int ileb;
	int sizeklms = sizeof(Klawisze) + 4 + 2;
	char *klawmsg = (char *) xmalloc(sizeklms);
	while ((ileb = recv(g->sockuchw, klawmsg, sizeklms, 0)) != -1) {
		if (ileb != sizeklms)
			continue;
		bcopy(klawmsg, &nrlud, 2);
		nrlud = ntohs(nrlud);
		if (nrlud >= g->ilelud)
			continue;
		bcopy(klawmsg + 2, &nrkom, 4);
		nrkom = ntohl(nrkom);
		if (nrkom > g->lud[nrlud]->nrostodb) {
			g->lud[nrlud]->nrostodb = nrkom;
			bcopy(klawmsg + 2 + 4, &(g->lud[nrlud]->klaw),
			      sizeof(Klawisze));
		}
	}
	free(klawmsg);
}

void gra_sock_init(StanGry * g, const char *inetadr, short port)
{
	struct sockaddr_in adres;
	struct in_addr inadres;

	g->sockuchw = socket(PF_INET, SOCK_DGRAM, 0);
	if (g->sockuchw == -1) {
		perror("socket()");
		exit(-1);
	}
	adres.sin_family = AF_INET;
	adres.sin_port = htons(port);
	if (!inet_aton(inetadr, &inadres)) {
		perror("inet_aton()");
		exit(-1);
	}
	adres.sin_addr = inadres;
	if (-1 ==
	    bind(g->sockuchw, (struct sockaddr *) &adres, sizeof(adres))) {
		perror("bind()");
		exit(-1);
	}
	if (-1 == fcntl(g->sockuchw, F_SETFL, O_NONBLOCK)) {
		perror("fcntl()");
		exit(-1);
	}
}

/*
 * Posuwa o 1 do przodu klatki animacji wybuchw.
 */
void gra_rusz_is(StanGry * g)
{
	int f;
	for (f = 0; f < g->ilepzlok; f++)
		if (g->przwsk[f][0].is < 0)
			g->przwsk[f][0].is++;
}

/*
 * Posuwa w przd znajdujce si w powietrzu pociski, sprawdza czy 
 * jaki gracz lub beczka zosta trafiony.
 */
void gra_rusz_strz(StanGry * g)
{
	int f;
	for (f = 0; f < g->ilemaxstrz; f++) {
		if (g->isstrz[f]) {
			int ff;
			for (ff = 0; ff < 6; ff++) {
				int fff;
				g->xstrz[f] +=
				    (int) (g->krokstrz * c[g->katstrz[f]]);
				g->ystrz[f] +=
				    (int) (g->krokstrz * s[g->katstrz[f]]);
				if (mapa[g->xstrz[f] / bok]
				    [g->ystrz[f] / bok] != 0) {
					g->isstrz[f] = 0;
					goto next;
				}
				for (fff = 0; fff < g->ilepzlok; fff++) {
					if ((g->xstrz[f] -
					     g->przwsk[fff][0].xpzl)
					    * (g->xstrz[f] -
					       g->przwsk[fff][0].xpzl) +
					    (g->ystrz[f] -
					     g->przwsk[fff][0].ypzl)
					    * (g->ystrz[f] -
					       g->przwsk[fff][0].ypzl)
					    < MINBS * MINBS
					    && g->przwsk[fff][0].is > 0) {
						g->przwsk[fff][0].is = -16;
						g->isstrz[f] = 0;
						goto next;
					}
				}
			}
		}
	      next:;
	}
}

/*
 * Sprawdza, czy gracz o numerze nrlud nie znalaz si zbyt blisko
 * ciany, jeli tak to odsuwa go.
 */
void gra_ods_od_scian(StanGry * g, int nrlud)
{
	int x = g->lud[nrlud]->pzl.xpzl;
	int y = g->lud[nrlud]->pzl.ypzl;

	if (x % bok < MINODL) {
		if (mapa[x / bok - 1][y / bok] != 0) {
			x = x - x % bok + MINODL;
		}
	}
	if (y % bok < MINODL)
		if (mapa[x / bok][y / bok - 1] != 0)
			y = y - y % bok + MINODL;
	if (bok - x % bok < MINODL)
		if (mapa[x / bok + 1][y / bok] != 0)
			x = x - x % bok + bok - MINODL;
	if (bok - y % bok < MINODL)
		if (mapa[x / bok][y / bok + 1] != 0)
			y = y - y % bok + bok - MINODL;
	if (y % bok < MINODL2 && x % bok < MINODL2)
		if (mapa[x / bok - 1][y / bok - 1] != 0) {
			if (y % bok < x % bok)
				x = x - x % bok + MINODL2;
			else
				y = y - y % bok + MINODL2;
		}
	if (y % bok < MINODL2 && bok - x % bok < MINODL2)
		if (mapa[x / bok + 1][y / bok - 1] != 0) {
			if (y % bok < bok - x % bok)
				x = x - x % bok + bok - MINODL2;
			else
				y = y - y % bok + MINODL2;
		}
	if (bok - y % bok < MINODL2 && bok - x % bok < MINODL2)
		if (mapa[x / bok + 1][y / bok + 1] != 0) {
			if (bok - y % bok < bok - x % bok)
				x = x - x % bok + bok - MINODL2;
			else
				y = y - y % bok + bok - MINODL2;
		}
	if (bok - y % bok < MINODL2 && x % bok < MINODL2)
		if (mapa[x / bok - 1][y / bok + 1] != 0) {
			if (bok - y % bok < x % bok)
				x = x - x % bok + MINODL2;
			else
				y = y - y % bok + bok - MINODL2;
		}
	g->lud[nrlud]->pzl.xpzl = x;
	g->lud[nrlud]->pzl.ypzl = y;
}

/*
 * Sprawdza, czy gracz o numerze nrlud nie znalaz si zbyt blisko beczki,
 * jeli tak to odsuwa go.
 */
void gra_ods_od_nier(StanGry * g, int nrlud)
{
	int x = g->lud[nrlud]->pzl.xpzl;
	int y = g->lud[nrlud]->pzl.ypzl;
	int f;
	for (f = 0; f < g->ilenieruchom; f++) {
		if (g->nieruchom[f][0].dist < MINBECZ
		    && g->nieruchom[f][0].is) {
			x = g->nieruchom[f][0].xpzl +
			    (int) ((x -
				    g->nieruchom[f][0].xpzl) * (MINBECZ /
								g->
								nieruchom
								[f][0].
								dist));
			y = g->nieruchom[f][0].ypzl +
			    (int) ((y -
				    g->nieruchom[f][0].ypzl) * (MINBECZ /
								g->
								nieruchom
								[f][0].
								dist));
		}
	}
	g->lud[nrlud]->pzl.xpzl = x;
	g->lud[nrlud]->pzl.ypzl = y;
}

/*
 * Oblicza odlegoci midzy graczem o numerze nrlud a pozostaymi
 * graczami i przedmiotami.
 */
void gra_ust_dist(StanGry * g, int nrlud)
{
	int x = g->lud[nrlud]->pzl.xpzl;
	int y = g->lud[nrlud]->pzl.ypzl;

	int f;
	for (f = 0; f < g->ilepzlok; f++)
		g->przwsk[f][0].dist =
		    sqrt((double)
			 ((x - g->przwsk[f][0].xpzl) * (x -
							g->przwsk[f][0].
							xpzl) + (y -
								 g->
								 przwsk[f]
								 [0].
								 ypzl) *
			  (y - g->przwsk[f][0].ypzl)));
}

/*
 * Inicjalizuje struktur gry na podstawie odczytanego pliku
 * konfiguracyjnego.
 */
void gra_ini(StanGry * g, Conf * cs)
{
	int f;
	g->oststart = 0;
	g->nrostnad = 0;
	g->nrostodb = 0;
	klawisze_ini(&(g->klaw));
	g->ilelud = cs->ilelud;
	g->ilepzlok = ilbecz + g->ilelud;
	g->przwsk =
	    (PrzZlok **) xmalloc(sizeof(PrzZlok *) * (g->ilepzlok));

	g->ilemaxstrz = 20;
	g->isstrz = (char *) xmalloc(sizeof(char) * (g->ilemaxstrz));
	g->xstrz = (int32_t *) xmalloc(sizeof(int32_t) * (g->ilemaxstrz));
	g->ystrz = (int32_t *) xmalloc(sizeof(int32_t) * (g->ilemaxstrz));
	g->katstrz =
	    (int32_t *) xmalloc(sizeof(int32_t) * (g->ilemaxstrz));
	g->krokstrz = 100;

	memset(g->isstrz, 0, g->ilemaxstrz);

	for (f = 0; f < ilbecz; f++) {
		g->przwsk[f] = (Beczka *) xmalloc(sizeof(Beczka));
		g->przwsk[f][0].xpzl = xbecz[f];
		g->przwsk[f][0].ypzl = ybecz[f];
		g->przwsk[f][0].is = 1;
	}

	g->ilenieruchom = ilbecz;
	g->nieruchom =
	    (PrzZlok **) xmalloc(sizeof(PrzZlok *) * (g->ilenieruchom));
	bcopy(g->przwsk, g->nieruchom,
	      g->ilenieruchom * sizeof(PrzZlok *));

	g->lud = (Ludzik **) xmalloc(sizeof(Ludzik *) * (g->ilelud));

	for (f = 0; f < g->ilelud; f++) {
		g->lud[f] = (Ludzik *) xmalloc(sizeof(Ludzik));
		ludzik_ini(g->lud[f]);

		g->lud[f]->pzl.is = 1;
		g->lud[f]->poza = 0;
		g->lud[f]->gunst = 0;
		g->lud[f]->gunready = 1;
		g->lud[f]->krok = 80;
		g->lud[f]->krokwbok = 25;
		g->lud[f]->katkrok = 10;
	}

	for (f = 0; f < g->ilelud; f++) {
		g->lud[f]->pzl.xpzl = cs->xs[f];
		g->lud[f]->pzl.ypzl = cs->ys[f];
		g->lud[f]->kierunek = cs->kier[f];
	}

	for (f = 0; f < g->ilelud; f++)
		g->przwsk[ilbecz + f] = &(g->lud[f]->pzl);
}

/*
 * W oparciu o odebrany przez sie stan klawiatury, wykonuje ruch gracza.
 */
void ludzik_ruch_klaw(Ludzik * l, StanGry * gra)
{
	int x = l->pzl.xpzl;
	int y = l->pzl.ypzl;
	int alfa = l->kierunek;

	if (l->klaw.rightstr) {
		x += (int) (l->krokwbok * s[(alfa + 240) % KATY]);
		y -= (int) (l->krokwbok * c[(alfa + 240) % KATY]);
	}
	if (l->klaw.leftstr) {
		x -= (int) (l->krokwbok * s[(alfa + 240) % KATY]);
		y += (int) (l->krokwbok * c[(alfa + 240) % KATY]);
	}
	if (l->klaw.left) {
		if (l->klaw.str) {
			x += (int) (l->krokwbok * s[(alfa + 240) % KATY]);
			y -= (int) (l->krokwbok * c[(alfa + 240) % KATY]);
		} else {
			alfa -= l->katkrok;
			if (alfa < 0)
				alfa += KATY;
		}
	}
	if (l->klaw.right) {
		if (l->klaw.str) {
			x -= (int) (l->krokwbok * s[(alfa + 240) % KATY]);
			y += (int) (l->krokwbok * c[(alfa + 240) % KATY]);
		} else
			alfa = (alfa + l->katkrok) % KATY;
	}
	if (l->klaw.up) {
		x += (int) (l->krok * c[(alfa + 240) % KATY]);
		y += (int) (l->krok * s[(alfa + 240) % KATY]);
		l->poza = !l->poza;
	}
	if (l->klaw.down) {
		x -= (int) (l->krok * c[(alfa + 240) % KATY]);
		y -= (int) (l->krok * s[(alfa + 240) % KATY]);
		l->poza = !l->poza;
	}
	if (!l->klaw.fire)
		l->gunready = 1;
	if (l->klaw.fire && !l->gunst && l->gunready) {
		int f;
		l->gunready = 0;
		l->gunst = 1;
		for (f = 0; f < 20; f++) {
			if (!gra->isstrz[f])
				break;
		}
		if (f == 20) {
			return;
		}
		gra->isstrz[f] = 1;
		gra->katstrz[f] = (alfa + KATY_8) % KATY;
		gra->xstrz[f] = x + (int) (MINBS * c[gra->katstrz[f]]);
		gra->ystrz[f] = y + (int) (MINBS * s[gra->katstrz[f]]);
	}
	l->kierunek = alfa;
	l->pzl.xpzl = x;
	l->pzl.ypzl = y;
}


int main()
{
	int f;
	Conf cs;
	StanGry gra;
	for (f = 0; f < KATY; f++) {
		s[f] = sin(f * PI / KATY_2);
		c[f] = cos(f * PI / KATY_2);
	}
	readmap("mapa/m2.mapa", "mapa/m2.pos");
	readconf(&cs, "config");
	gra_ini(&gra, &cs);
	gra_sock_init(&gra, cs.serwaddr, cs.serwport);
	for (f = 0; f < cs.ilelud; f++)
		wpisz_adr(&(gra.lud[f]->adres), cs.addrs[f], cs.ports[f]);

	do {
		gra_sluchaj_klaw(&gra);

		for (f = 0; f < gra.ilelud; f++) {
			if (gra.lud[f]->pzl.is == 0
			    && gra.lud[f]->klaw.fire)
				gra_fresh_start(&gra, f);
		}

		for (f = 0; f < gra.ilelud; f++) {
			if (gra.lud[f]->pzl.is <= 0)
				continue;
			ludzik_ruch_klaw(gra.lud[f], &gra);
			gra_ust_dist(&gra, f);
			gra_ods_od_nier(&gra, f);
			gra_ods_od_scian(&gra, f);
		}

		gra_rusz_strz(&gra);

		for (f = 0; f < gra.ilelud; f++)
			gra_slij_do(&gra, f);

		gra_rusz_is(&gra);

		for (f = 0; f < gra.ilelud; f++) {
			if (gra.lud[f]->pzl.is <= 0)
				continue;
			ludzik_rusz_gun(gra.lud[f]);
		}

		usleep(12000);
	}
	while (1);
}
